package highway;

import org.graphstream.graph.Edge;
import org.graphstream.graph.Graph;
import org.graphstream.graph.Node;
import org.graphstream.graph.implementations.SingleGraph;
import org.graphstream.stream.file.FileSource;
import org.graphstream.stream.file.FileSourceDGS;

import java.io.IOException;
import java.io.InputStream;
import java.util.LinkedList;
import java.util.Queue;

/**
 * @author Yoann Eichelberger
 * @author Vivien VAUDRY
 */
public class App {

    private Graph       network;
    private String      capacityAttribute = "cap";
    private Node        source, sink;
    private double      totalFlow;
    private Queue<Node> queue = new LinkedList<>();

    /**
     * Fonction principale, le programme s'éxecute ici.
     * 
     * @param args pas utilisé.
     */
    public static void main(String[] args) {
        System.setProperty("org.graphstream.ui.renderer", "org.graphstream.ui.j2dviewer.J2DGraphRenderer");
        Graph g = new SingleGraph("Highway");
        readResource("highway.dgs", g, new FileSourceDGS());
        g.addAttribute("ui.quality");
        g.addAttribute("ui.antialias");
        g.addAttribute("ui.stylesheet", "url('highway.css')");
        g.display(false);

        App mf = new App();
        mf.setCapacityAttribute("cap");
        mf.init(g);
        mf.setSource(g.getNode("A"));
        mf.setSink(g.getNode("I"));
        mf.compute();

        System.out.println(mf.getFlow());

        for (Node n : g.getEachNode())
            n.addAttribute("ui.label", "" + n.getId());

        for (Edge e : g.getEachEdge()) {
            double flow = mf.getFlow(e);
            double cap = mf.getCapacity(e);
            if (flow == 0) 
                e.addAttribute("ui.style", "fill-color: #999999;");
            if (flow > 0) 
                e.addAttribute("ui.label", "" + (int) flow + " (" + (int)cap + ")");
            // flow sature
            if (cap == flow) 
                e.addAttribute("ui.style", "fill-color: red;");
        }

        // g.getEdge("AC").addAttribute("ui.style", "fill-color: #009999;");
        g.getEdge("CD").addAttribute("ui.style", "fill-color: #009999;");
        // g.getEdge("DH").addAttribute("ui.style", "fill-color: #009999;");
        // g.getEdge("HI").addAttribute("ui.style", "fill-color: #009999;");
    }

    /**
     * Permet d'initialiser un graphe à partir d'une resource.
     * 
     * @param resourceName nom de la resource.
     * @param graph graphe à initialiser.
     * @param source source du fichier.
     */
    private static void readResource(String resourceName, Graph graph, FileSource source) {
        try (InputStream stream = App.class.getClassLoader().getResourceAsStream(resourceName)) {
            source.addSink(graph);
            source.readAll(stream);
        } catch (IOException ex) {
            ex.printStackTrace();
        } finally {
            source.removeSink(graph);
        }
    }


    /**
     * Créé un réseau à partir d'un graphe.
     * 
     * @param graph graphe à copier en réseau.
     */
    public void init(Graph graph) {
        this.network = new SingleGraph("network");
        for (Node n : graph) 
            this.network.addNode(n.getId());
        for (Edge e : graph.getEachEdge()) {
            Edge e1 = this.network.addEdge(e.getId(), e.getSourceNode().getId(), e.getTargetNode().getId(), true);
            e1.addAttribute("cap", e.hasNumber(this.capacityAttribute) ? e.getNumber(this.capacityAttribute) : Double.POSITIVE_INFINITY);
            e1.addAttribute("flow", 0);
            Edge e2 = this.network.addEdge(e.getId() + "_rev", e.getTargetNode().getId(), e.getSourceNode().getId(), true);
            e2.addAttribute("cap", 0);
            e2.addAttribute("flow", 0);
        }
    }

    /**
     * Construit le flot max du réseau.
     */
    public void compute() {
        this.totalFlow = 0;
        bfs();
        while (this.sink.hasAttribute("arc")) {
            augmentFlow();
            bfs();
        }
    }

    /**
     * Défini le noeud source du réseau.
     * 
     * @param n nouveau noeud source.
     */
    public void setSource(Node n) {
        this.source = this.network.getNode(n.getId());
    }

    /**
     * Défini le noeud évier du réseau.
     * 
     * @param n nouveau noeud évier.
     */
    public void setSink(Node n) {
        this.sink = this.network.getNode(n.getId());
    }

    /**
     * Défini l'attribut de la capacité.
     * 
     * @param capacityAttribute le nouvel attribut de la capacité.
     */
    public void setCapacityAttribute(String capacityAttribute) {
        this.capacityAttribute = capacityAttribute;
    }

    /**
     * Récupère le flot total.
     * 
     * @return le flot total.
     */
    public double getFlow() {
        return this.totalFlow;
    }

    /**
     * Récupère le flot d'un arc du réseau.
     * 
     * @param e un arc du réseau.
     * @return le flot d'un arc du réseau.
     */
    public double getFlow(Edge e) {
        return this.network.getEdge(e.getId()).getNumber("flow");
    }

    /**
     * Récupère la capacité maximal d'un arc du réseau.
     * 
     * @param e un arc du réseau.
     * @return la capacité maximal d'un arc du réseau.
     */
    public double getCapacity(Edge e) {
        return this.network.getEdge(e.getId()).getNumber("cap");
    }

    /**
     * 
     */
    private void bfs() {
        this.queue.clear();
        this.queue.add(this.source);
        for (Node n : this.network)
            n.removeAttribute("arc");
        while (!this.queue.isEmpty()) {
            Node n = this.queue.remove();
            for (Edge e : n.getEachLeavingEdge()) {
                double cap = e.getNumber("cap");
                double flow = e.getNumber("flow");
                Node neighbor = e.getTargetNode();
                if (neighbor != this.source && !neighbor.hasAttribute("arc") && cap > flow) {
                    neighbor.addAttribute("arc", e);
                    this.queue.add(neighbor);
                }
            }
        }
    }

    /**
     * 
     */
    private void augmentFlow() {
        double df = Double.POSITIVE_INFINITY;
        for (Edge e = this.sink.getAttribute("arc"); e != null; e = e.getSourceNode().getAttribute("arc"))
            df = Math.min(df, e.getNumber("cap") - e.getNumber("flow"));
        this.totalFlow += df;
        for (Edge e = this.sink.getAttribute("arc"); e != null; e = e.getSourceNode().getAttribute("arc")) {
            e.addAttribute("flow", e.getNumber("flow") + df);
            Edge er = e.getTargetNode().getEdgeToward((Node) e.getSourceNode());
            er.addAttribute("flow", er.getNumber("flow") - df);
        }
    }
}
